@@ -3,6 +3,7 @@ import random
 from typing import Tuple, List, Union
 from dh2vrml.util import rand_color
 import numpy as np
+from scipy.spatial.transform import Rotation
 from x3d.x3d import (
     X3D,
     Scene, Viewpoint, NavigationInfo,
@@ -54,7 +55,7 @@ def revolute_joint(color : Tuple[float, float, float]) -> Transform:
         ]
     )
 
-def prismatic_joint(color : Tuple[float, float, float]) -> Shape: 
+def prismatic_joint(color : Tuple[float, float, float]) -> Shape:
     return Shape(
         geometry=Box(
             size=(UL, UL, 2*UL)
@@ -79,7 +80,7 @@ def link_cross_section() -> List[Tuple[float, float]]:
     cross_section.append((radius*np.cos(0), radius*np.sin(0),))
     return cross_section
 
-def end_effector(color : Tuple[float, float, float]) -> Shape: 
+def end_effector(color : Tuple[float, float, float]) -> Shape:
     shaft_length = 1.5*UL
     gripper_width = 0.7*UL
     gripper_length = 0.7*UL
@@ -125,7 +126,7 @@ def end_effector(color : Tuple[float, float, float]) -> Shape:
         ]
     )
 
-def get_link_body(idx: int, d: float, theta: float, r: float, alpha: float, color: Tuple[float, float, float]) -> Transform:
+def get_link_body(idx: int, d: float, theta: float, r: float, alpha: float, offset: Tuple[float, float, float], last_offset: Tuple[float, float, float], color: Tuple[float, float, float], last_link=False) -> Transform:
     def get_z(d, theta):
         return np.matrix([
             [np.cos(theta), -np.sin(theta), 0, 0],
@@ -141,16 +142,32 @@ def get_link_body(idx: int, d: float, theta: float, r: float, alpha: float, colo
             [0, np.sin(alpha),  np.cos(alpha), 0],
             [0, 0            ,  0            , 1]
         ])
+    def get_offset_matrix(x, y, z, reverse=False):
+        if reverse:
+            x = -x
+            y = -y
+            z = -z
+        return np.matrix([
+            [1, 0, 0, x],
+            [0, 1, 0, y],
+            [0, 0, 1, z],
+            [0, 0, 0, 1],
+        ])
 
-    # Assume zero rotation in actuator, wrap geometry in a Transform frame with
-    # rotation at the end for simplicity
-    z = get_z(d, 0)
+    last_offset_matrix = get_offset_matrix(*last_offset, reverse=True)
+    z = get_z(d, theta)
     x = get_x(r, alpha)
-    t = z*x
+    offset_matrix = get_offset_matrix(*offset)
+    t = last_offset_matrix*z*x*offset_matrix
     zero_point = np.matrix([0, 0, 0, 1]).transpose()
+    x_dir = np.matrix([1, 0, 0, 0]).transpose()
     final_point = t*zero_point
+    final_x_dir = t*x_dir
+    # Link should always appear to connect into the side of a joint
+    second_last_point = final_point - UL*final_x_dir
     # Extract x, y, z from final point
     final_point = tuple(final_point[0:3].transpose().tolist()[0])
+    second_last_point = tuple(second_last_point[0:3].transpose().tolist()[0])
     extrusion_spine = [
         (0, 0, 0),  # Start extrusion at center of joint
         (0, 0, 1.5*PREV_UL),  # Extrusion protrudes up in the Z axis out of joint
@@ -161,11 +178,13 @@ def get_link_body(idx: int, d: float, theta: float, r: float, alpha: float, colo
             (PREV_UL, 0, 1.5*PREV_UL),
             (PREV_UL, 0, 0),
         ])
+    # Last link connects to the end effector, not a joint
+    if not last_link:
+        extrusion_spine.append(second_last_point)
     extrusion_spine.append(final_point)
 
     return Transform(
         DEF=f'l{idx}_link_body',
-        rotation=(0, 0, 1, theta),
         children=[
             Shape(
                 geometry=Extrusion(
@@ -192,21 +211,39 @@ def get_joint(type: JointType, color: Tuple[float, float, float]) -> _X3DChildNo
         # End effector should always be cyan
         return end_effector((0, 1, 1))
 
-def get_link(idx: int, d: float, theta: float, r: float, alpha: float, joint_type: JointType, last_joint_type: JointType, color: Union[Tuple[float, float, float], None]=None) -> Tuple[Transform, Transform]:
+def get_link(idx: int, d: float, theta: float, r: float, alpha: float, joint_type: JointType, last_joint_type: JointType, offset: Tuple[float, float, float], last_offset: Tuple[float, float, float], color: Union[Tuple[float, float, float], None]=None) -> Tuple[Transform, Transform]:
     if not color:
         color = rand_color()
+    if joint_type == JointType.END_EFFECTOR:
+        last_link = True
+    else:
+        last_link = False
     joint = get_joint(joint_type, color)
     link_alpha = Transform(
         DEF=f'l{idx}_alpha',
         rotation=(1, 0, 0, alpha),
         children=[
-            joint
+            Transform(
+                DEF=f'l{idx}_offset',
+                translation=offset,
+                children=[
+                    joint
+                ]
+            )
         ]
     )
     link = Transform(
         DEF=f'l{idx}_{last_joint_type.name}',
         children=[
-            get_link_body(idx, d, theta, r, alpha, color),
+            Transform(
+                DEF=f'l{idx}_link_offset',
+                translation=last_offset,
+                children=[
+                    get_link_body(
+                        idx, d, theta, r, alpha, offset, last_offset, color,
+                        last_link=last_link),
+                ]
+            ),
             Transform(
                 DEF=f'l{idx}_d',
                 translation=(0, 0, d),
@@ -230,9 +267,72 @@ def get_link(idx: int, d: float, theta: float, r: float, alpha: float, joint_typ
     )
     return link, link_alpha
 
+def get_viewpoint(
+        camera_location: Tuple[float, float, float],
+        camera_center: Tuple[float, float, float],
+        ) -> Viewpoint:
+    # From https://github.com/numpy/numpy/issues/5228
+    def cart2sph(x, y, z):
+        hxy = np.hypot(x, y)
+        r = np.hypot(hxy, z)
+        el = np.arctan2(z, hxy)
+        az = np.arctan2(y, x)
+        return az, el, r
+    camera_location_m = np.matrix(camera_location)
+    camera_center_m = np.matrix(camera_center)
+    desired_orientation = camera_center_m - camera_location_m
+    azimuth, inclination, _ = cart2sph(*desired_orientation.tolist()[0])
 
-def base_model(base_joint : JointType) -> X3D:
+    # Default orientation is straight down, we want to:
+    camera_rotvec = Rotation.from_euler(
+        'XYX',
+        [
+            np.pi/2, # Pan camera up to point towards positive Y in the fixed frame
+            (azimuth - np.pi/2), # Pan camera left towards camera center
+            inclination, # Pan camera up/down towards camera center
+        ]
+        ).as_rotvec()
+    camera_angle = np.linalg.norm(camera_rotvec)
+    camera_rotvec /= camera_angle
+    camera_orientation = tuple(camera_rotvec.tolist()) + (camera_angle,)
+
+    return Viewpoint(
+        position=camera_location,
+        centerOfRotation=camera_center,
+        orientation=camera_orientation
+    )
+
+
+def base_model(
+        base_joint: JointType,
+        base_offset: Tuple[float, float, float],
+        camera_location: Tuple[float, float, float],
+        camera_center: Tuple[float, float, float],
+        ) -> Tuple[X3D, Transform]:
     base_color = (0.2, 0.2, 0.2)
+    base = Transform(
+        DEF="Base Joint Offset",
+        translation=base_offset,
+        children=[
+            Transform(
+                DEF="Base",
+                translation=(0, 0, -1*UL),
+                children=[
+                    Shape(
+                        geometry=Box(
+                            size=(10, 10, 0.1)
+                        ),
+                        appearance=Appearance(
+                            material=Material(
+                                diffuseColor=base_color
+                            )
+                        )
+                    ),
+                ]
+            ),
+            get_joint(base_joint, base_color)
+        ]
+    )
     return X3D(
         profile="Immersive", version="3.3",
         Scene=Scene(
@@ -240,60 +340,50 @@ def base_model(base_joint : JointType) -> X3D:
                 NavigationInfo(
                     DEF="ExamineMode"
                 ),
-                Viewpoint(
-                    orientation=(-1, 0, 0, 0.2),
-                    position=(0, 5, 15),
-                ),
-                Transform(
-                    DEF="Base",
-                    translation=(0, 0, -1),
-                    children=[
-                        Shape(
-                            geometry=Box(
-                                size=(10, 10, 0.1)
-                            ),
-                            appearance=Appearance(
-                                material=Material(
-                                    diffuseColor=base_color
-                                )
-                            )
-                        ),
-                    ]
-                ),
-                get_joint(base_joint, base_color)
+                get_viewpoint(camera_location, camera_center),
+                base
             ]
         )
-    )
+    ), base
 
 
-def build_x3d(params : DhParams) -> X3D:
+def build_x3d(
+        params: DhParams,
+        camera_location: Union[Tuple[float, float, float], None]=(10, -10, 10),
+        camera_center: Union[Tuple[float, float, float], None]=(0, 0, 0)) -> X3D:
     global UL
     parameters = params.params
     scale = params.scale
     colors = params.colors
     joint_types = params.joint_types
+    offsets = params.offsets
     base_joint = joint_types.pop(0)
+    base_offset = offsets.pop(0)
 
     if scale[0] is not None:
         update_ul(scale[0])
     else:
         reset_ul()
     joint_types.append(JointType.END_EFFECTOR)
-    model = base_model(base_joint)
+    # Assume last frame is always centered around end effector
+    offsets.append((0, 0, 0))
+    model, ptr = base_model(base_joint, base_offset, camera_location, camera_center)
     ptr = model.Scene
 
     last_joint_type = base_joint
-    for idx, (p, j, c, s) in enumerate(zip(parameters, joint_types, colors, scale)):
+    last_offset = base_offset
+    for idx, (p, j, c, s, o) in enumerate(zip(parameters, joint_types, colors, scale, offsets)):
         if s is not None:
             update_ul(s)
         else:
             update_ul(UL)
         # link 0 is the base "link"
         link_idx = idx + 1
-        link, new_ptr = get_link(link_idx, p.d, p.theta, p.r, p.alpha, j, last_joint_type, c)
+        link, new_ptr = get_link(link_idx, p.d, p.theta, p.r, p.alpha, j, last_joint_type, o, last_offset, c)
         ptr.children.append(link)
         ptr = new_ptr
         last_joint_type = j
+        last_offset = o
 
     # Return UL to original length
     reset_ul()
